
from threading import Lock
from functools import wraps

class LockoutTagout:
    """Function decorator for simple mutual exclusions with tags.

    This decorator takes a tag as a parameter.
    Decorated functions with the same tag will acquire a lock
    associated with that tag before they are executed.
    For example, if functions f and g are both
    decorated with @Loto('data_x'), then they will
    execute with mutual exclusion. If function
    h is decorated with @Loto('data_y') then it will
    not acquire the same mutex as f and g and can
    run simultaneously.
    """
    
    locks = {}

    def __init__(self, tag):
        self.tag = tag
        if self.tag not in LockoutTagout.locks:
            LockoutTagout.locks[self.tag] = Lock()

    def __call__(self, wrappedFunc):
        # the decorator requires args that get passed to __init__
        # so the decorated function is passed here.
        # We acquire the lock associated with this tag before calling
        # the wrapped function
        @wraps(wrappedFunc)  # make the wrapped look like original
        def lockedCall(*args, **kwargs):
            with LockoutTagout.locks[self.tag]:
                wrappedFunc(*args, **kwargs)
        lockedCall._loto_tag = self.tag
        return lockedCall




import pytest
from src.loto import LockoutTagout
from time import sleep
from threading import Thread

# fixture for managing the LockoutTagout class variables
@pytest.fixture
def loto_fixture():
    LockoutTagout.locks.clear()

def test_tags_different_locks(loto_fixture):
    # Test if making 2 loto functions with different tags results
    # in different locks being made
    @LockoutTagout('tag1')
    def f():
        pass
    
    @LockoutTagout('tag2')
    def g():
        pass

    assert len(LockoutTagout.locks) == 2
    assert f._loto_tag is not g._loto_tag


def test_tags_same_lock(loto_fixture):
    # test that making 2 loto functions with the same tag
    # will associate the same lock with each
    @LockoutTagout('tag1')
    def f():
        pass

    @LockoutTagout('tag1')
    def g():
        pass
    assert len(LockoutTagout.locks) == 1
    assert f._loto_tag is g._loto_tag

def test_same_tags_locked_out(loto_fixture):
    # Make 2 threads. On lasts a long time but 
    # starts first. THe other thread tries to finish
    # immediately but starts after the first.
    # We test that the first thread actually finishes first
    # because they share a lock
    report = []

    @LockoutTagout('tag1')
    def slowFunc():
        sleep(1)
        report.append('slow')
        return

    @LockoutTagout('tag1')
    def fastFunc():
        report.append('fast')
        return

    slowThread = Thread(target=slowFunc)
    fastThread = Thread(target=fastFunc)

    slowThread.start()
    fastThread.start()

    fastThread.join()
    slowThread.join()

    assert report[0] == 'slow' and report[1] == 'fast'



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='loto',
    version='0.1.5',
    description='Decorator for tagging critical sections locking them out',
    long_description=long_description,
    description_content_type='text/markdown',
    long_description_content_type='text/markdown; charset=UTF-8; variant=GFM', 
    url='https://github.com/mardigontoler/loto',
    author='Mardigon Toler',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    packages=find_packages(),  # Required
    python_requires='>=3',
    install_requires=['pytest'],
)
